#!/usr/bin/env python
import os
import glob
import re
import shutil
import argparse
import subprocess as sp

import platform
if platform.system() == 'Windows':
    encoding = 'cp932'
else:
    encoding = 'utf-8'

# -pで指定できるパラメータ
PARA = {"SHIFTN": 4,
        "BKGD": 12,
        "SCALE": 1,
        "SCALE1": 1,
        "SCALE2": 1,
        "FWHM12": 4, "ASYM12": 4, "ETA12": 4, "ANISOBR12": 2, "PREF": 6, "CELLQ": 7,
        "FWHM122": 4, "ASYM122": 4, "ETA122": 4, "ANISOBR122": 2, "PREF1": 6, "CELLQ1": 7,
        "FWHM123": 4, "ASYM123": 4, "ETA123": 4, "ANISOBR123": 2, "PREF2": 6, "CELLQ2": 7}

description = """rietan run program."""
par = argparse.ArgumentParser(description=description)
par.add_argument('-i', '--insfile', default=None)
par.add_argument('-p', '--parameter', nargs='+', default=[])
par.add_argument('-l', '--atompara', nargs='+', default=[])
par.add_argument('-b', '--unbackup', default=False,  action='store_true')
par.add_argument('-u', '--unupdate', default=False,  action='store_true')
par.add_argument('-d', '--debug', default=False,  action='store_true')
par.add_argument('--Sthread', default=0.0,  type=float)
args = par.parse_args()


class RIETAN():

    def __init__(self, base):
        self.base, ext = os.path.splitext(args.insfile)
        self.body = body

    def UpdateVariable(self):
        def SetVal(str, val):
            exp = "^(\s*%s\s*=)\s*\d\s*([:\n])" % (str)
            obj = re.search(exp, self.body, re.MULTILINE)
            if obj is None:
                print("Variable %s is invalid." % (str))
                exit(1)
            new = "\\1 %d\\2" % (val)
            body = re.sub(exp, new, self.body, flags=re.M)
            return body
        # NUPDT = 1 (insファイルの更新しない)
        self.body = SetVal("NUPDT", not args.unupdate)
        # NPAT = 2 (itx形式でファイル保存)
        self.body = SetVal("NPAT", 2)
        # NBEAM = 1 (converntional XRD)
        self.body = SetVal("NBEAM", 1)
        # NMODE = 0 (0:Rietveld 1:Simulation 2:MEM)
        # self.body = SetVal("NMODE", 0)
        # NTARG = 4 (Cu K_alpha radiation.)
        self.body = SetVal("NTARG", 4)
        # NSURFR = 0 (Do not correct for surface roughness.)
        self.body = SetVal("NSURFR", 0)
        # NTRAN = 0 (Bragg-Brentano geometry (conventional divergence slit)
        self.body = SetVal("NTRAN", 0)
        # NPRFN = 2 (Modified split pseudo-Voigt function)
        self.body = SetVal("NPRFN", 2)
        # NLESQ = 2 (Marquardt method (recommended in most cases).
        self.body = SetVal("NLESQ", 0)
        # NAUTO = 2 (Refine incrementally (automatic; recommended in most
        # cases).
        self.body = SetVal("NAUTO", 2)

    def UpdateParam(self):
        params, val = [], []
        for i, p in enumerate(args.parameter):
            if i % 2 == 0:
                params.append(p)
            else:
                val.append(p)
        for i, p in enumerate(params):
            if p == "BKGD":
                restr = "(BKGD(\s+[\-0-9E\.]+){0,12}\s*\n(\s+[\-0-9E\.]+){0,12})\s+[01]{12}"
                if re.search(restr, self.body) is None:
                    print("BKGD was not match")
                    exit(1)
                self.body = re.sub(restr, "\\1  %s" % (val[i]), self.body)
            else:
                n = PARA[p]
                restr = "(%s(\s+[\-0-9E\.]+){%d})\s+[01]{%d}" % (p, n, n)
                if re.search(restr, self.body) is None:
                    print("%s was not match" % (p))
                    exit(1)
                self.body = re.sub(restr, "\\1  %s" % (val[i]), self.body)

    def UpdateAtomList(self):
        for i, p in enumerate(args.atompara):
            if i % 2 == 1:
                expr = "^(\s*%s\/\S+\s+([\+\-\.E0-9]+\s+){5}\s+)[012]{5}" % \
                    (args.atompara[i - 1])
                new = "\\1 %s" % (p)
                self.body = re.sub(expr, new, self.body, flags=re.M)

    def BackupIns(self):
        files = glob.glob(self.base + "-[0-9][0-9][0-9][0-9].ins")
        if len(files) == 0:
            num = 0
        else:
            number = []
            for f in files:
                o = re.search(self.base + "-(\d{4}).ins", f).group(1)
                number.append(int(o))
                num = max(number) + 1
        ext = "-%04d.ins" % (num)
        shutil.copy(self.base + ".ins", self.base + ext)

    def GetDataFromLst(self, lst, stderr):
        def Error():
            print("stdout:", lst)
            print("\n\nRIETAN PROGRAM WAS FAULT!!!", end=' ')
            print("Check %s.lst file." % (self.base))
            exit(1)
        lst_ = lst
        lst = lst.decode(encoding)
        # obj = re.search("^\s*\*\*\* End of job \*\*\*$", lst, re.M)
        obj = re.search("\s+\*\*\* End of job \*\*\*", lst)
        if stderr is not None or obj is None:
            Error()
        Rwobj = re.search("^\s*Rwp.*$", lst, re.M)
        Rbobj = re.search("^\s*RB.*$", lst, re.M)
        if args.debug is True:
            print(lst)
            print("------------------ debug ------------------")
            print("Rw: %s" % (Rwobj))
            print("Rb: %s" % (Rbobj))
            print("------------------ debug ------------------")
        if stderr is not None or Rwobj is None or Rbobj is None:
            Error()
        errlog = ["The coefficient matrix is not positive definite",
                  "Bad imaginary species line or phase line",
                  "Check decay parameters",
                  "Too many steps in SUBROUTINE QROMB"]
        for e in errlog:
            obj = re.search(e, lst, re.M)
            if obj is not None:
                Error()

        R = Rwobj.group(0)
        Rwp = re.search("Rwp\s*?=(\s*[0-9\.\-]+)", R).group(1)
        Re = re.search("Re\s*?=(\s*[0-9\.\-]+)", R).group(1)
        Re = re.search("Re\s*?=(\s*[0-9\.\-]+)", R).group(1)
        S = re.search("S\s*?=(\s*[0-9\.\-]+)", R).group(1)
        R = Rbobj.group(0)
        RB = re.search("RB\s*?=(\s*[0-9\.\-]+)", R).group(1)
        RF = re.search("RF\s*?=(\s*[0-9\.\-]+)", R).group(1)
        print("RIETAN was successfully finished.")
        print("-" * 70)
        print("Convergence parameters:")
        print("Rwp      = %s (<15)" % (Rwp))
        print("Re       = %s" % (Re))
        print("S=Rw/Re  = %s (<1.3)" % (S))
        print("RB       = %s (<10)" % (RB))
        print("RF       = %s (<10)" % (RF))
        print("-" * 70)
        lstfile = self.base + ".lst"
        o = open(lstfile, "wb")
        o.write(lst_)
        o.flush()
        return S

    def Run(self):
        base = self.base
        o = open(self.base + ".ins", "w")
        o.write(self.body)
        o.flush()
        o.close()
        com = """ rietan {base}.ins {base}.int {base}.bkg {base}.itx {base}.hkl
        {base}.xyz {base}.fos {base}.ffe {base}.fba {base}.ffi {base}.ffo
        {base}.vesta {base}.plt {base}.gpd {base}.alb {base}.prf {base}.inflip
        {base}.exp"""
        c = com.format(base=base).split()

        pp = sp.Popen(c, stdout=sp.PIPE, stdin=sp.PIPE, stderr=sp.STDOUT)
        stdout, stderr = pp.communicate()
        S = self.GetDataFromLst(stdout, stderr)
        print("%s.ins, %s.lst, and %s.itx were updated." %
              (self.base, self.base, self.base))
        print("-" * 70)
        print("Check the fitting pattern: itxplot.py %s.itx" % (self.base))
        S = float(S)
        if S < args.Sthread:
            print("\n")
            exit()
            print("run_rietan is stopped. (S: %f<%f)\n" %
                  (S, args.Sthread), end=' ')
            print("\n")
            exit(1)


def CheckArgumentParameter():
    if len(args.parameter) % 2 != 0:
        print("Parameter must be key [01]")
        exit(1)
    for i, a in enumerate(args.parameter):
        if i % 2 == 0:
            if (a in PARA) is False:
                print("Parameter %s is invalid." % (a))
                exit(1)
        else:
            if re.match("^[01]+$", a) is None or\
                    PARA[args.parameter[i - 1]] != len(a):
                print("Flags (%s) for %s is invalid." %
                      (a, args.parameter[i - 1]))
                print("Flags for %s must be %s [0|1]." %
                      (args.parameter[i - 1], PARA[args.parameter[i - 1]]))
                exit(1)


def CheckArgument():
    if args.insfile is None:
        insfile = glob.glob("*.int")
        if len(insfile) == 0:
            print("*.ins was not found in current directory.")
            exit(1)
        base, ext = os.path.splitext(insfile[0])
        args.insfile = base + ".ins"
    try:
        f = open(args.insfile, encoding=encoding)
    except IOError:
        print(('cannot open %s', args.infile))

    body = f.read()
    f.close()
    reex = "CELLQ\d{0,1}.+?\n(.+?)\}"
    o = re.search(reex, body, re.DOTALL)
    if o is None:
        print("labels can not be picked up.")
    labels = o.group(1)
    labels = re.sub("\s*[#!].*", "", labels)
    labels = re.sub("\s*$", "", labels, re.DOTALL)
    labels = re.sub("^\s*", "", labels, re.DOTALL)
    labels = labels.split("\n")
    for i, l in enumerate(labels):
        labels[i] = l.split()[0].split('/')[0]
    for i, a in enumerate(args.atompara):
        if i % 2 == 0:
            if (a in labels) is False:
                print("Atomic label %s is invalid." % (a))
                exit(1)
        else:
            if re.match("^[012]{5}$", a) is None:
                print("Flags (%s) for %s is invalid." %
                      (a, args.atompara[i - 1]))
                exit(1)
    return body


CheckArgumentParameter()
body = CheckArgument()
r = RIETAN(body)
r.UpdateParam()
r.UpdateVariable()
r.UpdateAtomList()
if not args.unbackup:
    r.BackupIns()
r.Run()
